
# THIS FILE IS GENERATED FROM OASYS SETUP.PY
short_version = '1.2.140'
version = '1.2.140'
full_version = '1.2.140'
git_revision = ''
release = True

if not release:
    version = full_version
    short_version += ".dev"
